/*
 * Copyright (c) 2004-2005, Dennis M. Sosnoski All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following
 * disclaimer. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 * following disclaimer in the documentation and/or other materials provided with the distribution. Neither the name of
 * JiBX nor the names of its contributors may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jibx.maven;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;

import org.apache.commons.io.FilenameUtils;
import org.apache.maven.artifact.Artifact;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.project.MavenProject;
import org.codehaus.plexus.util.DirectoryScanner;

/**
 * Runs the JiBX binding compiler.
 *
 * @author                        <a href="mailto:mail@andreasbrenk.com">Andreas Brenk</a>
 * @author                        <a href="mailto:frankm.os@gmail.gom">Frank Mena</a>
 * @author                        <a href="mailto:don@donandann.com">Don Corley</a>
 */
public abstract class AbstractJibxMojo extends AbstractMojo {

    //~ Static fields/initializers -------------------------------------------------------------------------------------

    static final String DEFAULT_INCLUDE_BINDINGS = "binding.xml";

    /**
     * The maven project.
     *
     * @resolveTransitiveDependencies
     * @parameter                      expression="${project}"
     * @required
     * @readonly
     */
    protected MavenProject project;

    /**
     * A list of modules to search for binding files in the format: groupID:artifactID
     *
     * @parameter  expression="${modules}"
     */
    HashSet<String> modules;

    /**
     * A list of modules or files to search for base binding files.
     * This can specify files in the local directory or files stored in your dependencies.
     *
     * If your based binding files are in a local file system, specify them as follows:
     * &lt;includeBaseBindings&gt;<br/>
     * &lt;includeBaseBinding&gt;<br/>
     * &lt;directory&gt;src/main/config&lt;/directory&gt;<br/>
     * &lt;includeBindings&gt;<br/>
     * &lt;includeBinding&gt;base-binding.xml&lt;/includeBinding&gt;<br/>
     * &lt;/includeBindings&gt;<br/>
     * &lt;/includeBaseBinding&gt;<br/>
     * &lt;/includeBaseBindings&gt;<br/>
     * 
     * If your based binding files are in a artifact that is one of your dependencies:
     * &lt;includeBaseBindings&gt;<br/>
     * &lt;includeBaseBinding&gt;<br/>
     * &lt;groupId&gt;com.mycompany.baseschema&lt;/groupId&gt;<br/>
     * &lt;artifactId&gt;base-schema&lt;/artifactId&gt;<br/>
     * &lt;classifier&gt;bindings&lt;/classifier&gt;<br/>
     * &lt;directory&gt;META-INF&lt;/directory&gt;<br/>
     * &lt;includeBindings&gt;<br/>
     * &lt;includeBinding&gt;base-binding.xml&lt;/includeBinding&gt;<br/>
     * &lt;/includeBindings&gt;<br/>
     * &lt;/includeBaseBinding&gt;<br/>
     * &lt;/includeBaseBindings&gt;<br/>
     * 
     * The classifier is optional (if your binding file is not in the main artifact)
     * A version is not necessary, since this declaration must be on your list of dependencies.
     * <b>Note: </b>For file filters, use the standard filter format described in the plexus
     * <a href="http://plexus.codehaus.org/plexus-utils/apidocs/org/codehaus/plexus/util/DirectoryScanner.html">DirectoryScanner</a>.<br/>
     * <b>Defaults value is:</b> binding.xml.
     * Include existing bindings and use mappings from the bindings for matching schema global definitions.
     * (this is the basis for modular code generation)
     * <b>Note:</b> If directory is not specified, relative paths start at &lt;directory&gt;.
     * 
     * @parameter  expression="${includeBaseBindings}"
     */
    HashSet<IncludeBaseBinding> includeBaseBindings;

    /**
     * Control flag multi-module mode.
     *
     * @parameter  expression="${multi-module}" default-value="false"
     * @required
     */
    boolean multimodule;

    /**
     * Exclude pattern for binding files.
     * @deprecated - This name was confusing since it is used as a binding file pattern
     * for bind and a schema file pattern for code-gen.
     * Use excludeBindings for binding and excludeSchema for schema
     * 
     * @parameter  expression="${excludes}"
     */
    @Deprecated
    ArrayList<String> excludes;

    /**
     * Include pattern for binding files.<br/>
     * @deprecated - This name was confusing since it is used as a binding file pattern
     * for bind and a schema file pattern for code-gen.
     * Use includeBindings for binding and includeSchema for schema
     * <b>Note: </b>Uses the standard filter format described in the plexus
     * <a href="http://plexus.codehaus.org/plexus-utils/apidocs/org/codehaus/plexus/util/DirectoryScanner.html">DirectoryScanner</a>.<br/>
     * <b>Defaults value is:</b> binding.xml.
     *
     * @parameter  expression="${includes}"
     */
    @Deprecated
    ArrayList<String> includes;

    /**
     * Exclude pattern for binding files.
     * 
     * @parameter  expression="${excludeSchemaBindings}"
     */
    ArrayList<String> excludeSchemaBindings;

    /**
     * Include pattern for binding files.<br/>
     * <b>Note: </b>Uses the standard filter format described in the plexus
     * <a href="http://plexus.codehaus.org/plexus-utils/apidocs/org/codehaus/plexus/util/DirectoryScanner.html">DirectoryScanner</a>.<br/>
     * <b>Defaults value is:</b> binding.xml.
     * Include existing bindings and use mappings from the bindings for matching schema global definitions.
     * (this is the basis for modular code generation)
     * Include base bindings as follows:<br/>
     * &lt;includeBindings&gt;<br/>
     * &lt;includeBinding&gt;base-binding.xml&lt;/includeBinding&gt;<br/>
     * &lt;/includeBindings&gt;<br/>
     * <b>Note:</b> Relative paths start at &lt;directory&gt;.
     *
     * @parameter  expression="${includeSchemaBindings}"
     */
    ArrayList<String> includeSchemaBindings;

    //~ Methods --------------------------------------------------------------------------------------------------------

    /**
     * Determines if running in single- or multi-module mode, collects all bindings and finally runs the binding
     * compiler.
     */
    public abstract void execute() throws MojoExecutionException, MojoFailureException;

    /**
     * Returns the basedir of the given project.
     */
    protected String getProjectBasedir(MavenProject project) {
        return FilenameUtils.normalize(project.getBasedir().getAbsolutePath());
    }

    /**
     * Verifies the plugins configuration and sets default values if needed.
     * Note: Remember to call inherited methods first.
     */
    protected void checkConfiguration() {
    	
    	if (this.includeSchemaBindings == null) {
            this.includeSchemaBindings = new ArrayList<String>();
        }
        if (this.excludeSchemaBindings == null) {
            this.excludeSchemaBindings = new ArrayList<String>();
        }

        if ((this.modules != null) && (this.modules.size() > 0)) {
            this.multimodule = true;
        } else {
            this.modules = null;
        }
        
        if (this.includeBaseBindings == null)
        	this.includeBaseBindings = new HashSet<IncludeBaseBinding>();
    }
    
    /**
     * Returns all bindings in the given directory according to the configured include/exclude patterns.
     */
    protected List<String> getIncludedFiles(String path, ArrayList<String> includeFiles, ArrayList<String> excludeFiles) throws MojoExecutionException, MojoFailureException {
        List<String> bindingSet = new ArrayList<String>();

        File bindingdir = new File(path);
        if (!bindingdir.exists())
        {	// Probably a url...
        	try {
				URL url = new URL(path);
				if ("file".equalsIgnoreCase(url.getProtocol()))
					return bindingSet;	// If you pass a file or directory it must exist, so I can scan it. 
			} catch (MalformedURLException e) {
				return bindingSet;
			}
			String[] includes = (String[]) includeFiles.toArray(new String[includeFiles.size()]);
			if ((includes != null) && (includes .length > 0))
			{
				for (String include : includes) {
					bindingSet.add(path + include);
				}
			}
			else
				bindingSet.add(path);	// Only supplied a single URL
        }
        if (!bindingdir.isDirectory()) {
            return bindingSet;
        }

        DirectoryScanner scanner = new DirectoryScanner();

        scanner.setBasedir(bindingdir);
        String[] includes = (String[]) includeFiles.toArray(new String[includeFiles.size()]);
        scanner.setIncludes(includes);
        String[] excludes = (String[]) excludeFiles.toArray(new String[excludeFiles.size()]);
        scanner.setExcludes(excludes);
        scanner.scan();

        String[] files = scanner.getIncludedFiles();
        String absolutePath = bindingdir.getAbsolutePath();
        for (int i = 0; i < files.length; i++) {
            String file = absolutePath + File.separator + files[i];
            bindingSet.add(file);
        }

        return bindingSet;
    }
    /**
     * Fix this file path.
     * If it is absolute, leave it alone, if it is relative prepend the default path or the base path.
     * @param filePath The path to fix
     * @param defaultPath The base path to use.
     * @return The absolute path to this file.
     */
    protected String fixFilePath(String filePath, String defaultPath)
    {
    	if (filePath == null)
    		return null;
        boolean relativePath = true;
        
        String basedir = defaultPath;
        if (basedir == null)
        	basedir = getProjectBasedir(this.project);
        if ((!filePath.startsWith(File.pathSeparator))
            	&& (!filePath.startsWith("/")))
            {	// Possible relative path
            		try {
                        File file = new File(basedir + File.separator + filePath);
                        if ((file.exists()) || (filePath.startsWith("target")))
                        	relativePath = false;	// If file exists, use absolute path
                    } catch (Exception e) {
                    	// Exception = use relative path
                    }
            }
        if (!relativePath)
        	filePath = basedir + File.separator + filePath;
        return filePath;
    }

    /**
     * Normalizes all entries.
     */
    /* package */ String[] normalizeClasspaths(Set<String> classpathSet) {
        String[] classpaths = (String[]) classpathSet.toArray(new String[classpathSet.size()]);

        for (int i = 0; i < classpaths.length; i++) {
            classpaths[i] = FilenameUtils.normalize(classpaths[i]);
        }

        return classpaths;
    }

    /**
     * Returns all bindings in the given directory according to the configured include/exclude patterns.
     */
    List<String> getBindings(String path) throws MojoExecutionException, MojoFailureException {
        return this.getIncludedFiles(path, this.includeSchemaBindings, this.excludeSchemaBindings);
    }

    /**
     * Returns all binding in the current project and all referenced projects (multi-module mode)
     */
    String[] getMultiModuleBindings() throws MojoExecutionException, MojoFailureException {
        Set<String> basedirSet = getProjectBasedirSet(this.project);
        Set<String> bindingSet = new HashSet<String>();
        // No Need to add project (single module mode) dir, it is included in the baseDirSet

        for (String basedir : basedirSet) {
            if (basedir.equals(getProjectBasedir(this.project)))
            	basedir = getFullPath(getBindingDirectory());	// Main project
            List<String> bindingList = getBindings(basedir);
            bindingSet.addAll(bindingList);
        }

        return (String[]) bindingSet.toArray(new String[bindingSet.size()]);
    }

    /**
     * Returns the classpath for the binding compiler running in multi-module mode.
     */
    String[] getMultiModuleClasspaths() throws MojoExecutionException, MojoFailureException {
        Set<String> classpathSet = getProjectCompileClasspathElementsSet(this.project);

        return normalizeClasspaths(classpathSet);
    }

    /**
     * Returns the basedir of the given project and all (or all in "modules" specified) reference projects.
     */
    @SuppressWarnings("unchecked")
	private Set<String> getProjectBasedirSet(MavenProject project) {
        Set<String> basedirSet = new HashSet<String>();
        basedirSet.add(getProjectBasedir(project));

        for (MavenProject projectReference : (Collection<MavenProject>)project.getProjectReferences().values()) {
            String projectId = projectReference.getGroupId() + ":" + projectReference.getArtifactId();

            if ((this.modules == null) || this.modules.contains(projectId)) {
                basedirSet.add(getProjectBasedir(projectReference));
            }
        }

        return basedirSet;
    }

    /**
     * Returns the build output directory of the given project.
     *
     * @throws  MojoExecutionException  if DependencyResolutionRequiredException occurs
     */
    protected abstract Set<String> getProjectCompileClasspathElements(MavenProject project) throws MojoExecutionException;

    /**
     * Returns the build output directory of the given project and all its reference projects.
     *
     * @throws  MojoExecutionException  if DependencyResolutionRequiredException occurs
     */
    @SuppressWarnings("unchecked")
	private Set<String> getProjectCompileClasspathElementsSet(MavenProject project) throws MojoExecutionException {
        Set<String> classpathElements = new HashSet<String>();
        classpathElements.addAll(getProjectCompileClasspathElements(project));

        for (MavenProject projectReference : (Collection<MavenProject>)project.getProjectReferences().values()) {
            classpathElements.addAll(getProjectCompileClasspathElements(projectReference));        	
        }

        return classpathElements;
    }

    /**
     * Returns all bindings in the current project (single-module mode).
     */
    String[] getSingleModuleBindings() throws MojoExecutionException, MojoFailureException {
        String bindingdir = getFullPath(getBindingDirectory());
        List<String> bindingSet = getBindings(bindingdir);

        return (String[]) bindingSet.toArray(new String[bindingSet.size()]);
    }

    /**
     * Returns the classpath for the binding compiler running in single-module mode.
     */
    String[] getSingleModuleClasspaths() throws MojoExecutionException, MojoFailureException {
        Set<String> classpathSet = getProjectCompileClasspathElements(this.project);

        return normalizeClasspaths(classpathSet);
    }

    /**
     * Get the binding path name for a includeBaseBinding binding.
     * This method actually unjars the binding file(s) from dependent resources.
     * @param basedir
     * @param includeBinding
     * @return
     */
	public String[] getBaseBindings(String[] bindings)
	{
    	List<String> bindingSet = new ArrayList<String>();
    	for (String binding : bindings) {
    		bindingSet.add(binding);
    	}
        if (includeBaseBindings.size() > 0)
    	{
	        for (IncludeBaseBinding includeDependentBinding : includeBaseBindings)
	        {
	            if (includeDependentBinding == null)
	            	continue;
				String groupId = includeDependentBinding.groupId;
				String artifactId = includeDependentBinding.artifactId;
				if ((groupId == null ) || (artifactId == null))
				{	// Probably a reference to a binding on the filesystem
					String directory = includeDependentBinding.directory;
					if (directory == null)
						directory = this.getBindingDirectory();
					String path = fixFilePath(directory, null);
			        File bindingdir = new File(path);
			        if (!bindingdir.exists())
			        {	// Probably a url...
			        	try {
							URL url = new URL(path);
							if (!"file".equalsIgnoreCase(url.getProtocol()))
							{
								String[] includes = (String[]) includeDependentBinding.includes.toArray(new String[includeDependentBinding.includes.size()]);
								if ((includes != null) && (includes .length > 0))
								{
									for (String include : includes) {
										bindingSet.add(path + include);
									}
								}
								else
									bindingSet.add(path);	// Only supplied a single URL
							}
							return bindingSet.toArray(new String[bindingSet.size()]);
						} catch (MalformedURLException e) {
							// Try as a file
						}
			        }

			        DirectoryScanner scanner = new DirectoryScanner();

			        scanner.setBasedir(bindingdir);
			        String[] includes = (String[]) includeDependentBinding.includes.toArray(new String[includeDependentBinding.includes.size()]);
			        scanner.setIncludes(includes);
			        //String[] excludes = (String[]) excludes.toArray(new String[excludeFiles.size()]);
			        //scanner.setExcludes(excludes);
			        for (String include : includes)
			        {
			        	if ((include.startsWith(File.separator))
			        		|| (include.startsWith("/")))
			        	{	// If the files specify an absolute path, don't scan
					        for (String includez : includes)
					        {
					            bindingSet.add(includez);
					        }
					        return bindingSet.toArray(new String[bindingSet.size()]);
			        	}
			        }
			        if (!bindingdir.isDirectory()) {
			            return bindingSet.toArray(new String[bindingSet.size()]);
			        }
			        scanner.scan();

			        String[] files = scanner.getIncludedFiles();
			        String absolutePath = bindingdir.getAbsolutePath();
			        for (int i = 0; i < files.length; i++) {
			            String file = absolutePath + File.separator + files[i];
			            bindingSet.add(file);
			        }
			        break;	// Done, got them!
				}
				@SuppressWarnings("unchecked")
				Set<Artifact> artifacts = this.project.getArtifacts();
		        for (Artifact artifact : artifacts)
		        {
		        	if ((!"jar".equals(artifact.getType()))
		        			&& (!"bundle".equals(artifact.getType()))
		        			&& (!"zip".equals(artifact.getType())))
		        		continue;
		        	if ((!groupId.equals(artifact.getGroupId()))
		        			|| (!artifactId.equals(artifact.getArtifactId())))
		        		continue;
		        	if (includeDependentBinding.classifier != null)
		        		if (!includeDependentBinding.classifier.equals(artifact.getClassifier()))
		        			continue;
		        	// Found it! Go through this jar and see if these binding files exist
					ZipFile zip = null;
					try {
						zip = new ZipFile(artifact.getFile());
					} catch (ZipException e) {
						e.printStackTrace();
					} catch (IOException e) {
						e.printStackTrace();
					}
					
					if ((includeDependentBinding.includes == null) || (includeDependentBinding.includes.size() == 0))
						bindingSet = extractBindingFile(bindingSet, zip, includeDependentBinding.directory, groupId, artifactId, DEFAULT_INCLUDE_BINDINGS);
					else
					{
						for (String include : includeDependentBinding.includes) {
							bindingSet = extractBindingFile(bindingSet, zip, includeDependentBinding.directory, groupId, artifactId, include);
						}
					}
		        }
	        }
        }

        return bindingSet.toArray(new String[bindingSet.size()]);
	}
	public List<String> extractBindingFile(List<String> pathList, ZipFile zip, String directory, String groupId, String artifactId, String filePathInJar)
	{
		try {
			String filenameIn = filePathInJar;
			if (directory != null)
			{
				if ((directory.endsWith(File.separator)) || (directory.endsWith("/")))
					filenameIn = directory + '/' + filenameIn;
				else
					filenameIn = directory + '/' + filenameIn;
			}
			ZipEntry entry = zip.getEntry(filenameIn);
			filenameIn = filePathInJar;
			int lastSlash = filenameIn.lastIndexOf(File.separator);
			if (lastSlash == -1)
				lastSlash = filenameIn.lastIndexOf('/');
			String startFilePath = "";
			if (lastSlash != -1)
			{
				startFilePath = File.separator + filenameIn.substring(0, lastSlash);
				filenameIn = filenameIn.substring(lastSlash + 1);
			}
			if (entry != null)
			{
				String filenameOut = getFullPath(getBindingDirectory()) + startFilePath;
				new File(filenameOut).mkdirs();
				filenameOut = filenameOut + File.separator + groupId + '-' + artifactId + '-' + filenameIn;
				File fileOut = new File(filenameOut);
				if (fileOut.exists())
					fileOut.delete();	// Hmmmmm - optimize LATER!
				InputStream inStream = zip.getInputStream(entry);
				if (fileOut.createNewFile())
				{
					OutputStream outStream = new FileOutputStream(filenameOut);
					byte[] b = new byte[1000];
					int size = 0;
					while ((size = inStream.read(b)) > 0)
					{
						outStream.write(b, 0, size);
					}
					outStream.close();
					inStream.close();
					pathList.add(filenameOut);
				}
			}
		
	} catch (IOException e) {
		e.printStackTrace();
	}
	return pathList;
}
    /**
     * Determine if the plugin is running in "multi-module" mode.
     */
    boolean isMultiModuleMode() {
        return this.multimodule;
    }

    /**
     * Determine if the plugin is running in "restricted multi-module" mode.
     */
    boolean isRestrictedMultiModuleMode() {
        return isMultiModuleMode() && (this.modules != null);
    }

    /**
     * Determine if the plugin is running in "single-module" mode.
     */
    boolean isSingleModuleMode() {
        return !isMultiModuleMode();
    }
    
    /**
     * Get the binding files directory.
     * @return The binding files directory.
     * @deprecated - Use getBindingDirectory
     */
    @Deprecated
    protected String getDirectory()
    {
    	return getBindingDirectory();
    }
    
    /**
     * Get the binding files directory.
     * @return The binding files directory.
     */
    protected abstract String getBindingDirectory();
    /**
     * Fix this directory path so it starts at the root project dir.
     * @param dir
     * @return
     */
    public String getFullPath(String dir)
    {
    	try {
			URL url = new URL(dir);
			if (!"file".equalsIgnoreCase(url.getProtocol()))
				return dir;	// If you pass a valid non file URL, use it!
		} catch (MalformedURLException e) {
		}
    	if ((!dir.startsWith(File.separator)) && (!dir.startsWith("/")))
    		dir = this.getProjectBasedir(project) + File.separator + dir;
    	return dir;
    }
}
